# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/61_tens.ipynb.

# %% auto 0
__all__ = ['ptdevice', 'pttensor', 'devchain', 'asdev', 'torchtype', 'gettype', 'try2dtype', 'mpsdtype', 'to', 'dtypechain',
           'tenchain', 'ten2cpu', 'ten2mps', 'ten2cuda', 'try2move', 'try2ten', 'asten', 'totorch', 'asnans',
           'torchnans', 'nanslike', 'assize', 'resize', 'resize_tensor']

# %% ../nbs/61_tens.ipynb 6
from functools import wraps

# %% ../nbs/61_tens.ipynb 8
from types import ModuleType
from typing import Any, overload

# %% ../nbs/61_tens.ipynb 9
#| export


# %% ../nbs/61_tens.ipynb 11
#| export


# %% ../nbs/61_tens.ipynb 13
try: import torch
except ImportError: ...

# %% ../nbs/61_tens.ipynb 15
try: import pytorch_lightning as pl
except ImportError: ...

# %% ../nbs/61_tens.ipynb 17
from nlit import TO, CPU, MPS, CUDA, DEVICE, FLOAT32, FLOAT64
from nlit import __DOC__, __NAME__, __MODULE__, __ANNOTATIONS__
from quac import (
    tensor, tensorq,
    device, deviceq, devcpu, devmps, devcuda,
    pttype, pttypeq, memfmt, presfmt, boolq, 
    nantensor as NanTensor, size as Size
)
from quac.eggs import pt, tv_functional
from chck import isnone, notnone, isdevice, isdtype, istensor
from sigr import applyfns, attrfunc

# %% ../nbs/61_tens.ipynb 19
from .cons import WRAPS_ASSIGN_ANNDOCS
from .atyp import P
from .arrs import asnp
from .bend import hasmps

# %% ../nbs/61_tens.ipynb 22
def ptdevice(a: Any, *args: P.args, **kwargs: P.kwargs) -> device: 
    '''Wrapper for `torch.device(a)`'''
    return torch.device(a)

def pttensor(a: Any, *args: P.args, **kwargs: P.kwargs) -> tensor: 
    '''Wrapper for `torch.as_tensor(a, ...)`'''
    return torch.as_tensor(a, *args, **kwargs)

# %% ../nbs/61_tens.ipynb 25
def try2dev(d: deviceq) -> device:
    '''Attempts to convert an input to a PyTorch device, falling back to the input if conversion fails.
    
    Parameters
    ----------
    d : str | device | None
        The input to convert.
        
    Returns
    -------
    device
        The PyTorch device or the original input if conversion is not possible.
    '''
    try: return d if isdevice(d) else ptdevice(d)
    except: return d

# %% ../nbs/61_tens.ipynb 27
def devchain(x: deviceq, funcs, *args: P.args, **kwargs: P.kwargs) -> device:
    '''Applies a sequence of functions (`funcs`) to an initial value `x`, 
    optionally stopping when the typeguard function `isdevice` returns `True`.
    
    See Also
    --------
    sigr.applyfns
    ''' 
    return applyfns(x, funcs, *args, check=isdevice, **kwargs)

# %% ../nbs/61_tens.ipynb 28
def asdev(d: deviceq, *args, **kwargs) -> device:
    '''Ensures an input is converted to a torch.device, applying a series of potential conversions.'''
    if isdevice(d): return d
    return devchain(d, (try2dev,), *args, **kwargs)

# %% ../nbs/61_tens.ipynb 30
def _dev_in_sig(dev: str | device, *args, **kwargs) -> bool:
    '''Checks if a device is specified in the arguments or keyword arguments of a function call.
    
    Parameters
    ----------
    dev : str | device
        The device to check for.
        
    *args : Any
        Positional arguments of the function call.
        
    **kwargs : Any
        Keyword arguments of the function call.
        
    Returns
    -------
    bool
        True if the device is specified, False otherwise.
        
    Examples
    --------
    >>> args = ('cpu', 1, 2, 'banana', )
    ... kwds = dict(device=device(type='mps'))
    
    >>> _dev_in_sig('mps', *args, **kwds)
    True
    
    >>> _dev_in_sig('cpu', *args, **kwds)
    True
    
    >>> _dev_in_sig('cuda', *args))
    False
    '''
    _dev = kwargs.get(dev, kwargs.get(DEVICE, None))
    if notnone(_dev) or dev in args or asdev(dev) in args: return True 
    return False

# %% ../nbs/61_tens.ipynb 32
def torchtype(dtype: str) -> str | pttype:
    '''Attempts to resolve a string to a PyTorch dtype, falling back to the input string if not found.
    
    Parameters
    ----------
    dtype : str
        The dtype string to resolve.
        
    Returns
    -------
    str | torch.dtype
        The PyTorch dtype if found, otherwise the input string.
        
    Examples
    --------
    >>> torchtype('float32'), torchtype('float64')
    (torch.float32, torch.float64)
    
    See Also
    --------
    gettype: Alias for torchtype
    '''
    try: return getattr(pt.duck, dtype)
    except: ...
    try: 
        import torch
        return getattr(torch, dtype)
    except: ...
    return dtype

@wraps(torchtype, assigned=WRAPS_ASSIGN_ANNDOCS)
def gettype(dtype: str) -> str | pttype:
    '''Alias for torchtype'''
    return torchtype(dtype)

# %% ../nbs/61_tens.ipynb 34
def try2dtype(a: tensor, dtype: pttype) -> tensor:
    '''Attempts to convert a tensor to the specified dtype, applying conversions as needed.
    
    Parameters
    ----------
    a : tensor
        The tensor to convert.
    
    dtype : pttype
        The target dtype.
        
    Returns
    -------
    tensor
        The tensor converted to the specified dtype, or the original tensor if conversion is not possible.
    '''
    try: return a.to(dtype=dtype) if istensor(a) else pttensor(asnp(a), dtype=dtype)
    except: return a

# %% ../nbs/61_tens.ipynb 36
def mpsdtype(dtype: pttype, device: deviceq = None, assume_on_mps: boolq = True) -> pttype:
    '''Adjusts the dtype for mps compatibility, converting float64 to float32 on mps devices.
    
    Coerces `dtype` to `torch.float32` if `dtype` is `torch.float64` and `device` is `'mps'`.
    
    Parameters
    ----------    
    dtype : pttype
        The original dtype to check against
    
    device : Device, default: None
        The device of the tensor or model from which the `dtype` comes from. If provided
        will be used to detemine whether or not to make `torch.float64`, `torch.float32`
        only if the device is actually `'mps'`.

    assume_on_mps: bool, default: True
        Whether or not to assume that the device of choice is `'mps'`. Setting this to
        `True` will result in `dtype` of `torch.float64` being converted to `torch.float32`
        to try and silently fix MPS errors.

    Notes
    -----
    - `float64` is not availble on `'mps'`, so they are coerced to `float32`
    - `assume_on_mps = True` and `hasmps() == True` are both needed as device might not 
    be provided.

    Returns
    -------
    dtype : torch.dtype
        the dtype, corrected for mps if needed
        
    Examples
    --------
    >>> mpsdtype(float), mpsdtype(torch.float), mpsdtype(torch.float64)
    (torch.float32, torch.float32, torch.float32)
    
    >>> mpsdtype(dtype = torch.float64, device = torch.device('mps'), assume_on_mps = False)
    torch.float32
    
    >>> to(torch.tensor([1,2,3]), 'mps')
    tensor([1, 2, 3], device='mps:0')
        
    See Also
    --------
    utrc.bend.hasmps : checks if mps is available
    '''
    float32: pttype = torchtype(FLOAT32)
    float64: pttype = torchtype(FLOAT64)
    is_float64 = (dtype == float64) or (dtype == FLOAT64) or (dtype == float)
    
    if notnone(device):
        if not isdevice(device): 
            device = asdev(device)
            
        if isinstance(device, str):
            is_device_mps = device == MPS
        else:
            is_device_mps = device.type == MPS
        if is_device_mps:
            assume_on_mps = True

        elif device.type == CUDA:
            assume_on_mps = False

    if hasmps() and assume_on_mps and is_float64: 
        return float32
    return dtype

# %% ../nbs/61_tens.ipynb 41
@overload
def to(x, other: tensor, nonblocking: bool = False, copy: bool = False) -> tensor: ...
@overload
def to(x, dtype, nonblocking: bool = False, copy: bool = False, memory_format: memfmt = presfmt) -> tensor: ...
@overload
def to(x, device: device, dtype: pttype = None, nonblocking: bool = False, copy: bool = False, memory_format: memfmt = presfmt) -> tensor: ...
# @attrfunc(TO)
def to(x, *args: P.args, **kwargs: P.kwargs) -> tensor:
    '''Wrapper for `tensor.to(...)` to transfers a tensor to a specified device 
    or changes its dtype or memory format.
    
    Parameters
    ----------
    x : tensor
        The tensor to manipulate.
        
    *args : P.args
        Positional arguments specifying the target device, dtype, or memory format.
        
    **kwargs : P.kwargs
        Keyword arguments for non-blocking and copy operations.
    
    Returns
    -------
    tensor
        The manipulated tensor.
    '''
    fn = getattr(x, TO, lambda *args, **kwargs: pttensor(x))
    has_mps = _dev_in_sig(MPS, *args, **kwargs)
    try:
        if istensor(x) and has_mps:
            dtype = mpsdtype(x.dtype, x.device)
            tokws = True
            args = list(args)
            for i, arg in enumerate(args):
                if isdtype(arg):
                    args[i] = dtype
                    tokws = False
                    break
            if tokws:
                kwargs.update(dtype=dtype)
            
        return fn(*args, **kwargs)
    except: return x

# %% ../nbs/61_tens.ipynb 46
def dtypechain(x: pttype, funcs, *args: P.args, **kwargs: P.kwargs) -> pttype: 
    '''Applies a sequence of functions (`funcs`) to an initial value `x`, 
    optionally stopping when the typeguard function `isdtype` returns `True`.
    
    See Also
    --------
    sigr.applyfns
    ''' 
    return applyfns(x, funcs, *args, check=isdtype, **kwargs)

# %% ../nbs/61_tens.ipynb 47
def tenchain(x: tensorq, funcs, *args: P.args, **kwargs: P.kwargs) -> tensor:
    '''Applies a sequence of functions (`funcs`) to an initial value `x`, 
    optionally stopping when the typeguard function `istensor` returns `True`.
    
    See Also
    --------
    sigr.applyfns
    ''' 
    return applyfns(x, funcs, *args, check=istensor, **kwargs)

# %% ../nbs/61_tens.ipynb 50
@wraps(to, assigned = (__MODULE__, __DOC__, __ANNOTATIONS__))
def ten2cpu(t: tensor, *args, **kwargs) -> tensor[devcpu]:
    '''Given a tensor, ensures that it is on cpu.
    
    Parameters
    ----------
    t : tensor
        The tensor to move.
    *args : Any
        Additional arguments for the `to` function.
    **kwargs : Any
        Additional keyword arguments for the `to` function.
        
    Returns
    -------
    tensor[devcpu]
        The tensor moved to the cpu.
    '''
    return to(t, device = asdev(CPU))

# %% ../nbs/61_tens.ipynb 52
@wraps(to, assigned = (__MODULE__, __DOC__, __ANNOTATIONS__))
def ten2mps(t: tensor, *args, **kwargs) -> tensor[devmps]:
    '''Ensures a tensor is moved to mps, adjusting dtype if needed.
    
    Given a tensor, ensures that it is on mac silicon.    
    Additionally makes sure `tensor` is `torch.float32` if  `tensor.dtype` is `torch.float64`.
    
    Parameters
    ----------
    t : tensor
        The tensor to move and potentially adjust dtype for mps compatibility.
    *args : Any
        Additional arguments for the `to` function.
    **kwargs : Any
        Additional keyword arguments for the `to` function.
        
    Returns
    -------
    tensor[devmps]
        The tensor moved to mps, with dtype adjusted if necessary.
        
    Examples
    --------
    >>> ten2mps(torch.as_tensor([1,2,3], dtype=torch.float64, device='cpu'))
    tensor([1., 2., 3.], device='mps:0')
    '''
    if not istensor(t): return t
    dtype = mpsdtype(t.dtype, (device := asdev(MPS)), assume_on_mps = False)
    return to(t, dtype = dtype, device = device)

# %% ../nbs/61_tens.ipynb 54
@wraps(to, assigned = (__MODULE__, __DOC__, __ANNOTATIONS__))
def ten2cuda(t: tensor, *args, **kwargs) -> tensor[devcuda]:
    '''Ensures a tensor is moved to cuda.
    
    Parameters
    ----------
    t : tensor
        The tensor to move.
    *args : Any
        Additional arguments for the `to` function.
    **kwargs : Any
        Additional keyword arguments for the `to` function.
        
    Returns
    -------
    tensor[devcuda]
        The tensor moved to cuda.
    '''
    return to(t, device = asdev(CUDA))

# %% ../nbs/61_tens.ipynb 56
@wraps(to, assigned = (__MODULE__, __DOC__, __ANNOTATIONS__))
def try2move(t: tensor, *args: P.args, **kwargs: P.kwargs) -> tensor:
    '''Attempts to move a tensor to another device or change its dtype 
    based on another tensor's properties if another tensor is provided.
    
    Parameters
    ----------
    t : tensor
        The tensor to move or convert.
        
    *args : Any
        Additional arguments, potentially including a target tensor to mimic.
        
    **kwargs : Any
        Additional keyword arguments for the `to` function.
        
    Other Parameters
    ----------------
    o : tensor, optional
        The tensor to mimic when moving or converting `t`.
        Note, it should be the first argument in `*args` if provided.
        
        
    Returns
    -------
    tensor
        The moved or converted tensor.
    '''
    if len(args) >= 1 and istensor(o := args[0]):
        dtype = mpsdtype(o.dtype, o.device, assume_on_mps = False)
        t = to(t, dtype=dtype)
    return to(t, *args, **kwargs)

# %% ../nbs/61_tens.ipynb 60
def try2ten(a: tensorq, dtype: pttypeq = None, device: deviceq = None) -> tensor:
    '''Attempts to convert any input to a PyTorch tensor, specifying dtype and device if given.
    
    Parameters
    ----------
    a : tensor | nparray | npmatrix | sparray | spmatrix | list | tuple | anndata
        The input to convert.
    
    dtype : pttype, optional
        The target dtype for the tensor.
        
    device : device, optional
        The target device for the tensor.
        
    Returns
    -------
    tensor
        The resulting PyTorch tensor.
    '''
    try: return a if istensor(a) else pttensor(asnp(a), dtype=dtype, device=device)
    except: return a

# %% ../nbs/61_tens.ipynb 62
def asten(a, *args, **kwargs) -> tensor:
    '''Converts an input to a PyTorch tensor, applying device and dtype adjustments as specified.
    
    Parameters
    ----------
    a : tensor | nparray | npmatrix | sparray | spmatrix | list | tuple | anndata
        The input to convert.
    
    *args : Any
        Additional arguments for conversion and adjustment functions.
        
    **kwargs : Any
        Additional keyword arguments for conversion and adjustment functions.
        
    Other Parameters
    ----------------
    dtype : pttype, optional
        The target dtype for the tensor.
        
    device : device, optional
        The target device for the tensor.
        
    Returns
    -------
    tensor
        The resulting PyTorch tensor.
        
    Examples
    --------
    >>> (
    ...    float64 := torchtype(FLOAT64),
    ...    a := asten([1,2,3]), 
    ...    b := asten([1,2,3], dtype=float64), 
    ...    x := asten([1,2,3], dtype=float64, device=CPU), 
    ...    y := asten([1,2,3], dtype=float64, device=CPU, mps=True),
    ...    z := asten([1,2,3], dtype=float64, mps=True)
    ... )
    (
    ...    torch.float64,
    ...    tensor([1, 2, 3]),                         # a (on cpu)
    ...    tensor([1., 2., 3.], dtype=torch.float64), # b (on cpu)
    ...    tensor([1., 2., 3.], dtype=torch.float64), # x (on cpu)
    ...    tensor([1., 2., 3.], device='mps:0'),      # y (on cpu but moved to mps)
    ...    tensor([1., 2., 3.], device='mps:0')       # z (on mps)
    )
    
    >>> to(a, z), asten(a, z)
    (tensor([1., 2., 3.], device='mps:0'), tensor([1., 2., 3.], device='mps:0'))
    
    >>> try2move(a, z) # move a (cpu) to z (mps)
    tensor([1., 2., 3.], device='mps:0')
    '''
    if istensor(a):
        return applyfns(a, (to, ), *args, **kwargs)

    chain = (asnp, try2move, try2ten, to, )
    a = tenchain(a, chain, *args, **kwargs)
    
    if _dev_in_sig(MPS, *args, **kwargs): 
        chain += (ten2mps, )
    
    if _dev_in_sig(CPU, *args, **kwargs): 
        chain += (ten2cpu, )
    
    if _dev_in_sig(CUDA, *args, **kwargs): 
        chain += (ten2cuda, )
    
    return applyfns(a, chain, *args, **kwargs)

@wraps(asten, assigned = (__MODULE__, __DOC__, __ANNOTATIONS__))
def totorch(a, *args, **kwargs) -> tensor:
    '''Alias for `asten`.'''
    return asten(a, *args, **kwargs)

# %% ../nbs/61_tens.ipynb 67
def asnans(size: tuple[int, ...]) -> NanTensor:
    '''
    Create a tensor of `size` filled with `torch.nan`

    Parameters
    ----------
    size : tuple[int, ...]
        The size of the tensor to be created.

    Returns
    -------
    NanTensor
        A tensor of the specified size filled with NaN values.
    '''
    return torch.nan * torch.zeros(size)

@wraps(asnans)
def torchnans(size: tuple) -> NanTensor:
    '''Alias for asnans'''
    return asnans(size)


def nanslike(t: tensor) -> NanTensor: 
    '''
    Create a tensor of the same size as the given tensor, filled with NaN values.

    Parameters
    ----------
    t : Tensor
        The tensor to match the size of.

    Returns
    -------
    NanTensor
        A tensor of the same size as `t`, filled with NaN values.
    '''
    return torchnans(t.shape)

# %% ../nbs/61_tens.ipynb 69
def assize(t: tensor, size: Size | None = None) -> tensor:
    '''Resizes a tensor

    Parameters
    ----------
    t : Tensor
        The input tensor to resize.
        
    resize : Size, optional
        The target size for resizing, does nothing if None.
        
    Returns
    -------
    Tensor
        The resized tensor.
        
    See Also
    --------
    asto.tens.assize : Alias for assize
    asto.tens.resize : Alias for assize
    asto.tens.resize_tensor : Alias for assize
    torchvision.transforms.functional.resize: Alias for assize
    '''
    if isnone(size): return t 
    F: ModuleType = tv_functional.duck
    t = F.resize(t, size)
    return t

@wraps(assize, assigned=WRAPS_ASSIGN_ANNDOCS)
def resize(t: tensor, size: Size | None = None) -> tensor:
    '''Alias for `assize`'''
    return assize(t, size=size)

@wraps(assize, assigned=WRAPS_ASSIGN_ANNDOCS)
def resize_tensor(t: tensor, size: Size | None = None) -> tensor:
    '''Alias for `assize`'''
    return assize(t, size=size)


# %% ../nbs/61_tens.ipynb 71
#| export



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/06_path.ipynb.

# %% auto 0
__all__ = ['pathchain', 'try2path', 'path2str', 'aspath']

# %% ../nbs/06_path.ipynb 6
import os
from functools import wraps

try: from pathlib import Path
except: ...

# %% ../nbs/06_path.ipynb 8
from typing import TypeGuard

# %% ../nbs/06_path.ipynb 10
#| export


# %% ../nbs/06_path.ipynb 12
from nchr import DOT
from nlit import __DOC__, __NAME__, __MODULE__, __ANNOTATIONS__
from nlit import (PATH, __FSPATH__, )
from quac import plibpath, pathlike, boolq
from chck import ispath, ispathlike
from sigr import applyfns

# %% ../nbs/06_path.ipynb 14
from .atyp import P

# %% ../nbs/06_path.ipynb 17
def pathchain(x, funcs, *args: P.args, **kwargs: P.kwargs): 
    '''Applies a sequence of functions (`funcs`) to an initial value `x`, 
    optionally stopping when the typeguard function `isarng` returns `True`.
    
    See Also
    --------
    sigr.applyfns
    ''' 
    return applyfns(x, funcs, *args, check=ispathlike, **kwargs)

# %% ../nbs/06_path.ipynb 20
def try2path(a) -> plibpath:
    '''Attempts to convert `a` to a numpy array representing an arithmetic range starting at 0.
    
    Parameters
    ----------
    a : Any
        The input object to convert.
        
    Returns
    -------
    np.array
        An arithmetic range numpy array if possible, otherwise `a` itself.
    '''
    if ispath(a): return a
    try: return Path(a)
    except: return a

# %% ../nbs/06_path.ipynb 22
def path2str(path: pathlike, failsafe: boolq = True) -> str:
    if issubclass(type(path), os.PathLike): return str(path.__fspath__())
    elif hasattr(path, __FSPATH__):         return str(path.__fspath__())
    elif hasattr(path, PATH):               return path2str(path.path)
    elif isinstance(path, str):             return path
    elif failsafe:                          return DOT
    else: raise ValueError(f'path2str: path {path} not recognized')

# %% ../nbs/06_path.ipynb 24
def aspath(path) -> Path:
    return Path(path2str(path))

def aspath(a, *args: P.args, **kwargs: P.kwargs) -> Path:
    if ispath(a): return a
    return Path(pathchain(a, (path2str, ), *args, **kwargs))

# %% ../nbs/06_path.ipynb 26
#| export




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/70_chns.ipynb.

# %% auto 0
__all__ = ['match_channels_format', 'ascfmt', 'as_channels_format', 'to_channels_format', 'aschnfmt', 'format_channels', 'aschns',
           'channels_resize', 'aschannels']

# %% ../nbs/70_chns.ipynb 6
from functools import wraps

# %% ../nbs/70_chns.ipynb 8
#| export


# %% ../nbs/70_chns.ipynb 10
try: import numpy as np
except: ...

# %% ../nbs/70_chns.ipynb 12
from quac import (tensor, int4, intq, size as Size)
from chck import isnone, isstr, istuple

# %% ../nbs/70_chns.ipynb 14
from asto.cons import (
    THWC, TCHW, THWC_AXES, TCHW_AXES,
    WRAPS_ASSIGN_ANNDOCS
)
from .atyp import P, Channels
from .tens import resize_tensor

# %% ../nbs/70_chns.ipynb 18
def _flip_channels_format(t: tensor) -> tensor:
    '''Flips the tensor between THWC and TCHW channel configurations.
    
    See Also
    --------
    etrc.enum.Channels.flip: Alias for _flip_channels_format
    '''
    return t.transpose(1, 3)

# %% ../nbs/70_chns.ipynb 20
def _guess_channels_index(t: tensor) -> int:
    '''Returns the index of the channel dimension
    
    See Also
    --------
    etrc.enum.Channels.cidx: Alias for _guess_channels_index
    '''
    idx = -1
    for i in range(len(t.shape)):
        if t.shape[i] != 3: 
            continue
        idx = i
    idx = idx if idx >= 0 else np.argmin(t.shape)
    if (val := t.shape[idx]) not in {1, 3}: 
        raise ValueError(f'Expected eitehr 1 or 3 channels, got {val}')
    return idx

# %% ../nbs/70_chns.ipynb 23
def _match_channels_format_string(fmt: str) -> str:
    match fmt.casefold():
        case THWC.casefold():
            return THWC
        case TCHW.casefold(): 
            return TCHW
        case _: 
            return THWC

# %% ../nbs/70_chns.ipynb 25
def _match_channels_format_tuple(fmt: int4) -> str:
    match fmt:
        case v if fmt == THWC_AXES: 
            return THWC
        case v if fmt == TCHW_AXES: 
            return TCHW
        case _: 
            return TCHW

# %% ../nbs/70_chns.ipynb 27
def match_channels_format(fmt: str | int4) -> str:
    '''Safely retrieves the channel format string, defaulting to 'THWC' if invalid.

    Parameters
    ----------
    fmt : Channels | 'thcw' | 'tchw' | (0, 1, 2, 3) | (0, 3, 1, 2)
        The channel configuration to validate.

    Returns
    -------
    Channels
        The validated channel configuration, defaults to THWC if input is invalid.
        
    See Also
    --------
    etrc.enum.Channels.safe: Alias for match_channels_format
    '''
    match fmt:
        case t if isstr(fmt): 
            return _match_channels_format_string(fmt)
        case t if istuple(fmt): 
            return _match_channels_format_tuple(fmt)
        case _: 
            return THWC

# %% ../nbs/70_chns.ipynb 29
def ascfmt(
    t: tensor, 
    fmt: Channels = THWC, 
    cidx: intq = None,
    *args: P.args,
    **kwargs: P.kwargs, 
) -> tensor:
    '''
    Converts a tensor to the specified channel configuration.

    Parameters
    ----------
    t : tensor
        The input tensor to convert.
    
    fmt : Channels | 'thcw' | 'tchw' | (0, 1, 2, 3) | (0, 3, 1, 2), default: 'thwc'
        The desired channel configuration to return.
        
    cidx : int, optional
        The current channel index in the tensor, calculated if not provided.
        
    Returns
    -------
    tensor
        The tensor converted to the specified channel configuration.
        
    See Also
    --------
    aschnfmt : Alias for ascfmt
    as_channels_format : Alias for ascfmt
    format_channels : Alias for ascfmt
    to_channels_format : Alias for ascfmt 
    etrc.enum.Channels.to : Alias for ascfmt
    '''
    # match the provided format to a valid channel configuration
    fmt = match_channels_format(fmt)
    # if channel_index is not provided, guess which axis belongs to channels 
    if isnone(cidx): cidx = _guess_channels_index(t)
    # flip the channels axis if needed
    match (fmt, cidx):
        case (THWC, 1): return _flip_channels_format(t)
        case (THWC, 3): return t
        case (TCHW, 1): return t
        case (TCHW, 3): return _flip_channels_format(t)
        case _: return t
    # raise ValueError(f'Invalid channel configuration: {fmt}, {channel_index}')

@wraps(ascfmt, assigned=WRAPS_ASSIGN_ANNDOCS)
def as_channels_format(t: tensor, fmt: Channels = THWC, cidx: intq = None, *args: P.args, **kwargs: P.kwargs) -> tensor:
    return ascfmt(t, fmt, cidx)

@wraps(ascfmt, assigned=WRAPS_ASSIGN_ANNDOCS)
def to_channels_format(t: tensor, fmt: Channels = THWC, cidx: intq = None, *args: P.args, **kwargs: P.kwargs) -> tensor:
    return ascfmt(t, fmt, cidx)

@wraps(ascfmt, assigned=WRAPS_ASSIGN_ANNDOCS)
def aschnfmt(t: tensor, fmt: Channels = THWC, cidx: intq = None, *args: P.args, **kwargs: P.kwargs) -> tensor:
    return ascfmt(t, fmt, cidx)

@wraps(ascfmt, assigned=WRAPS_ASSIGN_ANNDOCS)
def format_channels(t: tensor, fmt: Channels = THWC, cidx: intq = None, *args: P.args, **kwargs: P.kwargs) -> tensor:
    return ascfmt(t, fmt, cidx)

# %% ../nbs/70_chns.ipynb 31
def aschns(
    t: tensor, 
    size: Size | None = None, 
    fmt: Channels = THWC, 
    *args: P.args, 
    **kwargs: P.kwargs, 
) -> tensor:
    '''Converts a tensor to the specified channel configuration adjusting the size
    as needed.

    Parameters
    ----------
    t : tensor
        The input tensor to convert.
            
    size : tuple[int, ...], optional
        The desired size of the output tensor.
        
    fmt : Channels | 'thcw' | 'tchw' | (0, 1, 2, 3) | (0, 3, 1, 2), default: 'thwc'
        The desired channel format to return.
    
    Returns
    -------
    tensor
        The tensor converted to the specified channel configuration.
        
    See Also
    --------
    aschannels : Alias for aschns
    channels_resize : Alias for aschns
    etrc.enum.Channels.resize : Alias for aschns
    '''
    if isnone(size): return as_channels_format(t, fmt=fmt)
    # Cast to TCHW format for compatiability with `torchvision.transforms.functional.resize`
    t = as_channels_format(t, fmt=TCHW)
    # Then we apply the `resize` operation
    t = resize_tensor(t, size)
    # Then we convert back to the desired format
    t = as_channels_format(t, fmt=fmt)
    return t

@wraps(aschns, assigned=WRAPS_ASSIGN_ANNDOCS)
def channels_resize(t: tensor, size: Size | None = None, fmt: Channels = THWC, *args: P.args, **kwargs: P.kwargs) -> tensor:
    return aschns(t, size=size, fmt=fmt, *args, **kwargs)

@wraps(aschns, assigned=WRAPS_ASSIGN_ANNDOCS)
def aschannels(t: tensor, size: Size | None = None, fmt: Channels = THWC, *args: P.args, **kwargs: P.kwargs) -> tensor:
    return aschns(t, size=size, fmt=fmt, *args, **kwargs)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_atyp.ipynb.

# %% auto 0
__all__ = ['T', 'O', 'P', 'Ts', 'CoordFormat', 'SliceFormat', 'SliceFormats', 'SliceTuple', 'CoordTuple', 'SliceSpec',
           'SliceSpecs', 'THWCFormat', 'TCHWFormat', 'ChannelsFormats', 'Channels']

# %% ../nbs/01_atyp.ipynb 6
#| export


# %% ../nbs/01_atyp.ipynb 8
from typing import Union, Literal, TypeVar, ParamSpec, TypeAlias, TypeVarTuple, ForwardRef, Iterable

# %% ../nbs/01_atyp.ipynb 10
#| export


# %% ../nbs/01_atyp.ipynb 12
from quac import int4, slicespec, slicelike

# %% ../nbs/01_atyp.ipynb 14
#| export


# %% ../nbs/01_atyp.ipynb 16
T = TypeVar('T')
'''Type variable for generic types''';

O = TypeVar('O')
'''Type variable for generic types reprenesting "other"''';

P = ParamSpec('P')
'''Parameter specification type variable, capturing the type of function parameters.''';

Ts = TypeVarTuple('Ts')

# %% ../nbs/01_atyp.ipynb 19
CoordFormat: TypeAlias = Literal['coord'];

SliceFormat: TypeAlias = Literal['slice'];

SliceFormats: TypeAlias = Union[CoordFormat, SliceFormat]

SliceTuple: TypeAlias = tuple[int | None, int, int | None]
'''Slice format tuple for a `slice`''';

CoordTuple: TypeAlias = tuple[int, ...]
'''Coordinate format tuple for a `slice`''';

SliceSpec: TypeAlias = slice | slicelike | slicespec | SliceTuple | CoordTuple
'''`type[slice] | type[slicelike] | type[slicespec] | type[tuple[int | None, int, int | None]] | type[tuple[int, ...]]`''';

SliceSpecs: TypeAlias = Iterable[SliceSpec]
'''`type[Iterable[slice | slicelike | slicespec | tuple[int | None, int, int | None] | tuple[int, ...]]]`''';

# %% ../nbs/01_atyp.ipynb 21
THWCFormat: TypeAlias = Literal['thwc'];
'''The time, height, width, and channels axis order.''';

TCHWFormat: TypeAlias = Literal['tchw'];
'''The time, channels, height, and width axis order.''';

ChannelsFormats: TypeAlias = Union[THWCFormat, TCHWFormat]

Channels: TypeAlias = Union[ChannelsFormats, str, tuple[int, int, int, int], int4, ForwardRef('Channels', module='etrc.enum', is_class=True)]
'''`Channels | 'thcw' | 'tchw' | (0, 1, 2, 3) | (0, 3, 1, 2)`''';

# %% ../nbs/01_atyp.ipynb 23
#| export



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_cons.ipynb.

# %% auto 0
__all__ = ['WRAPS_ASSIGN_ANNDOCS', 'COORD', 'SLICE', 'THWC', 'TCHW', 'TIME_HEIGHT_WIDTH_CHANNELS', 'TIME_CHANNELS_HEIGHT_WIDTH',
           'THWC_AXES', 'TCHW_AXES', 'TIME_HEIGHT_WIDTH_CHANNELS_AXES', 'TIME_CHANNELS_HEIGHT_WIDTH_AXES']

# %% ../nbs/00_cons.ipynb 6
#| export


# %% ../nbs/00_cons.ipynb 8
from typing import TypeVar, ParamSpec, TypeAlias, TypeVarTuple

# %% ../nbs/00_cons.ipynb 10
#| export


# %% ../nbs/00_cons.ipynb 12
from nlit import (
    TCHW, THWC, COORD, SLICE, 
    __DOC__, __MODULE__, __ANNOTATIONS__
)

# %% ../nbs/00_cons.ipynb 14
#| export


# %% ../nbs/00_cons.ipynb 16
WRAPS_ASSIGN_ANNDOCS = (__DOC__, __MODULE__, __ANNOTATIONS__)

# %% ../nbs/00_cons.ipynb 18
COORD = COORD
'''Coordinate format for slice specifications''';

SLICE = SLICE
'''Slice format for slice specifications''';

# %% ../nbs/00_cons.ipynb 20
THWC = THWC
'''The time, height, width, and channels axis order.''';

TCHW = TCHW
'''The time, channels, height, and width axis order.''';

TIME_HEIGHT_WIDTH_CHANNELS = THWC
'''The time, height, width, and channels axis order.''';

TIME_CHANNELS_HEIGHT_WIDTH = TCHW
'''The time, channels, height, and width axis order.''';

THWC_AXES = (0, 1, 2, 3)
'''The time, height, width, and channels axis order as a tuple of integers.''';

TCHW_AXES = (0, 3, 1, 2)
'''The time, channels, height, and width axis order as a tuple of integers.''';

TIME_HEIGHT_WIDTH_CHANNELS_AXES = THWC_AXES
'''The time, height, width, and channels axis order as a tuple of integers.''';

TIME_CHANNELS_HEIGHT_WIDTH_AXES = TCHW_AXES
'''The time, channels, height, and width axis order as a tuple of integers.''';


# %% ../nbs/00_cons.ipynb 22
#| export



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/31_arrs.ipynb.

# %% auto 0
__all__ = ['detach', 'clone', 'cpu', 'numpy', 'todense', 'tolist', 'toarray', 'arrchain', 'ann2arr', 'asnp', 'asarr', 'as01']

# %% ../nbs/31_arrs.ipynb 6
from functools import wraps

# %% ../nbs/31_arrs.ipynb 8
from typing import Any, Union, TypeAlias, ParamSpec

# %% ../nbs/31_arrs.ipynb 10
try: import numpy as np
except ModuleNotFoundError: np = None

# %% ../nbs/31_arrs.ipynb 12
from nlit import (
    CPU, TOARRAY, TODENSE, TOLIST, DETACH, CLONE, NUMPY, LAYERS,
    __DOC__, __NAME__, __MODULE__, __ANNOTATIONS__
)
from chck import isad, isarr, issparse, istrc, isnpmatrix
from sigr import attrfunc, applyfns
from quac import (strq, nparray, npmatrix, tensor, adata, sparray, spmatrix, devcpu)

# %% ../nbs/31_arrs.ipynb 14
from .atyp import P

# %% ../nbs/31_arrs.ipynb 17
def ndarray(a: Any, *args: P.args, **kwargs: P.kwargs) -> nparray:
    '''Wrapper for `np.array(a, ...)`'''
    return np.array(a, *args, **kwargs)

# %% ../nbs/31_arrs.ipynb 19
@attrfunc(DETACH)
def detach(x: tensor) -> tensor: '''try: x.detach() except: x'''

@attrfunc(CLONE)
def clone(x: tensor) -> tensor: '''try: x.clone() except: x'''

@attrfunc(CPU)
def cpu(x: tensor) -> tensor[devcpu]: '''try: x.cpu() except: x'''

@attrfunc(NUMPY)
def numpy(x: tensor) -> nparray: '''try: x.numpy() except: x'''

@attrfunc(TODENSE)
def todense(x: tensor) -> nparray: '''try: x.todense() except: x'''

@attrfunc(TOLIST)
def tolist(x: tensor) -> list: '''try: x.tolist() except: x'''

@attrfunc(TOARRAY)
def toarray(x: tensor) -> nparray: '''try: x.toarray() except: x'''

# %% ../nbs/31_arrs.ipynb 21
def arrchain(x, funcs, *args: P.args, **kwargs: P.kwargs):
    '''Applies a sequence of functions (`funcs`) to an initial value `x`, 
    optionally stopping when the typeguard function `isarr` returns `True`.
    
    See Also
    --------
    sigr.applyfns
    ''' 
    return applyfns(x, funcs, *args, check=isarr, **kwargs)

# %% ../nbs/31_arrs.ipynb 22
def try2arr(a) -> nparray:
    '''Attempts to convert an input to an ndarray, falling back to the input if conversion fails.
    
    Parameters
    ----------
    a : nparray | npmatrix | sparray | spmatrix | tensor | anndata | list | tuple 
        The input to convert.
        
    Returns
    -------
    np.array
        The resulting numpy array.
    
    '''
    try: return a if isarr(a) else ndarray(a)
    except: return a

# %% ../nbs/31_arrs.ipynb 24
def trc2arr(a: tensor) -> nparray:
    '''torch.tensor --> np.array
    
    Converts a PyTorch tensor to a NumPy array by detaching, cloning, moving to CPU, and converting.
    
    Parameters
    ----------
    a : tensor
        The tensor to convert.
        
    Returns
    -------
    np.array
        The resulting array.

    Examples
    --------
    >>> t = torch.tensor([...])
    ... t.detach().clone().cpu().numpy()
    np.array([...])
    '''
    if isarr(a) or not istrc(a): return a
    return arrchain(a, (detach, clone, cpu, numpy, ))

# %% ../nbs/31_arrs.ipynb 26
def mtx2arr(a: npmatrix) -> nparray:
    '''np.matrix --> np.array
    
    Converts an np.matrix to an np.array using the `tolist` and `ndarray` methods.
    
    Parameters
    ----------
    a : npmatrix
        The numpy.matrix object to convert from.
        
    Returns
    -------
    np.array
        The resulting array.
    
    Examples
    --------
    >>> m = np.matrix([...])
    ... np.array(m.tolist())
    np.array([...])
    '''
    if isarr(a) or not isnpmatrix(a): return a
    return arrchain(a, (tolist, ndarray, ))

# %% ../nbs/31_arrs.ipynb 28
def spr2arr(a: Union[spmatrix, sparray]) -> nparray:
    '''scipy.sparse --> np.array
    
    Converts a scipy.sparse matrix to an np.array using the `toarray` method.
    
    Parameters
    ----------
    a : spmatrix | sparray
        The scipy sparse object to convert.
        
    Returns
    -------
    np.array
        The resulting array.
    
    Examples
    --------
    >>> m = scipy.sparse.csr_matrix([...])
    ... m.toarray()
    np.array([...])
    '''
    if isarr(a) or not issparse(a): return a
    return arrchain(a, (toarray, ))

# %% ../nbs/31_arrs.ipynb 30
def dns2arr(a: Union[spmatrix, sparray]) -> nparray:
    '''scipy.sparse --> np.array
    
    Converts a scipy.sparse matrix to a dense np.array using the `todense` and `tolist` methods.
    
    Parameters
    ----------
    a : spmatrix | sparray
        The scipy sparse object to convert.
        
    Returns
    -------
    np.array
        The resulting dense array.
    
    Examples
    --------
    >>> m = scipy.sparse.csr_matrix([...])
    ... np.array(m.todense().tolist())
    np.array([...])
    '''
    if isarr(a) or not issparse(a): return a
    return arrchain(a, (todense, tolist, ndarray, ))

# %% ../nbs/31_arrs.ipynb 32
def ann2arr(a: adata, layer: strq = None) -> nparray:
    '''anndata.AnnData --> np.array
    
    Extracts an array from an AnnData object, optionally from a specified layer.

    Parameters
    ----------
    a : adata
        The AnnData object to extract from.
    
    layer : str, optional
        The layer to extract from. If None, the X attribute is used.
        
    Returns
    -------
    np.array
        The extracted array.

    Examples
    --------
    >>> a = anndata.AnnData(X=[...], ...)
    ... a.X
    np.array([...])
    
    >>> a = anndata.AnnData(X=[...], ..., layers=dict(...))
    >>> a.layers('layer', a.X)
    np.array([...])
    '''
    if isarr(a) or not isad(a) or not hasattr(a, LAYERS): return a
    arr = a.layers.get(layer, a.X)
    return arrchain(arr, (toarray, ndarray))

# %% ../nbs/31_arrs.ipynb 34
def obj2arr(a, *args: P.args, **kwargs: P.kwargs) -> nparray:
    '''object --> np.array
    
    Attempts to convert any object to an ndarray using a series of conversion functions.
    
    Parameters
    ----------
    a : nparray | npmatrix | sparray | spmatrix | tensor | anndata | list | tuple 
        The object to convert.
    *args : Any
        Additional arguments passed to conversion functions.
    **kwargs : Any
        Additional keyword arguments passed to conversion functions.
        
    Other Parameters
    ----------------
    layer : str, optional
        The layer to extract from an AnnData object if `a` is an AnnData object.
        
    Returns
    -------
    nparray
        The resulting ndarray, or the original object if all conversions fail.
        
    Examples
    --------
    >>> import numpy as np, torch
    ... ten = torch.tensor(np.random.randint(0, 5, size=(2, 3)))
    ... ten, asnp(ten)
    (
    ... tensor([[4, 4, 3],
    ...         [3, 2, 1]]),
    ... array([[4, 4, 3],
    ...        [3, 2, 1]])
    )
    
    >>> asnp(np.matrix([1,2,3]))
    array([[1, 2, 3]])
    
    >>> # invoke spr2arr
    ... a = scipy.sparse.csr_matrix([...])
    ... a.toarray()
    np.array([...])
    
    >>> # invoke dns2arr
    ... a = scipy.sparse.csr_matrix([...])
    ... np.array(a.todense().tolist())
    np.array([...])
    
    >>> # invoke trc2arr
    ... a = torch.tensor([...])
    ... a.detach().clone().cpu().numpy()
    np.array([...])
    
    >>> # invoke mtx2arr
    ... a = np.matrix([...])
    ... np.array(a.tolist())
    np.array([...])
    
    >>> # invoke ann2arr
    ... a = anndata.AnnData(X=[...], ...)
    ... a.X
    np.array([...])
    
    >>> # invoke try2arr
    ... a = [...]
    ... np.array(a)
    '''
    if isarr(a): return a
    return arrchain(a, (
            spr2arr, # `scipy.sparse.csr_matrix([...]).toarray()`
            trc2arr, # `torch.tensor([...]).detach().clone().cpu().numpy()`
            dns2arr, # `np.array(scipy.sparse.csr_matrix([...]).todense().tolist())`
            mtx2arr, # `np.array(np.matrix([...]).tolist())`
            ann2arr, # `AnnData([...]).layers.get(layer, a.X).toarray()`
            try2arr, # `np.array`
        ), *args, **kwargs
    )

# %% ../nbs/31_arrs.ipynb 35
@wraps(obj2arr)
def mth2arr(a, *args: P.args, **kwargs: P.kwargs) -> nparray:
    '''Wrapper for obj2arr to maintain backwards compatibility with the `mth2arr` function.'''
    return obj2arr(a, *args, **kwargs)

# %% ../nbs/31_arrs.ipynb 37
@wraps(obj2arr, assigned=(__DOC__, __MODULE__, __ANNOTATIONS__))
def asnp(a, *args, **kwargs) -> nparray: 
    '''Ensures an input is converted to an np.array, applying a series of potential conversions.'''
    return obj2arr(a, *args, **kwargs)

# %% ../nbs/31_arrs.ipynb 38
@wraps(asnp)
def asarr(a, *args: P.args, **kwargs: P.kwargs) -> nparray:
    '''Wrapper for asnp to maintain backwards compatibility with the `asarr` function.'''
    return obj2arr(a, *args, **kwargs)

# %% ../nbs/31_arrs.ipynb 43
def as01(a: list, cutoff: float = 0.5) -> nparray:
    '''Binarizes an input array based on a cutoff threshold.
    
    Parameters
    ----------
    a : list
        The input list to binarize.
        
    cutoff : float, default: 0.5
        The threshold above which values are set to 1, and below to 0. 
        Default is 0.5.
        
    Returns
    -------
    nparray
        The binarized array.
    '''
    return (asnp(a) > cutoff).astype(int)



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/asto',
                'doc_host': 'https://dsm-72.github.io',
                'git_url': 'https://github.com/dsm-72/asto',
                'lib_path': 'asto'},
  'syms': { 'asto.arng': { 'asto.arng._safe_idx': ('arng.html#_safe_idx', 'asto/arng.py'),
                           'asto.arng._safe_len': ('arng.html#_safe_len', 'asto/arng.py'),
                           'asto.arng.arange': ('arng.html#arange', 'asto/arng.py'),
                           'asto.arng.arngchain': ('arng.html#arngchain', 'asto/arng.py'),
                           'asto.arng.arr2arng': ('arng.html#arr2arng', 'asto/arng.py'),
                           'asto.arng.asarng': ('arng.html#asarng', 'asto/arng.py'),
                           'asto.arng.df2arng': ('arng.html#df2arng', 'asto/arng.py'),
                           'asto.arng.isarng': ('arng.html#isarng', 'asto/arng.py'),
                           'asto.arng.try2arng': ('arng.html#try2arng', 'asto/arng.py')},
            'asto.arrs': { 'asto.arrs.ann2arr': ('arrs.html#ann2arr', 'asto/arrs.py'),
                           'asto.arrs.arrchain': ('arrs.html#arrchain', 'asto/arrs.py'),
                           'asto.arrs.as01': ('arrs.html#as01', 'asto/arrs.py'),
                           'asto.arrs.asarr': ('arrs.html#asarr', 'asto/arrs.py'),
                           'asto.arrs.asnp': ('arrs.html#asnp', 'asto/arrs.py'),
                           'asto.arrs.clone': ('arrs.html#clone', 'asto/arrs.py'),
                           'asto.arrs.cpu': ('arrs.html#cpu', 'asto/arrs.py'),
                           'asto.arrs.detach': ('arrs.html#detach', 'asto/arrs.py'),
                           'asto.arrs.dns2arr': ('arrs.html#dns2arr', 'asto/arrs.py'),
                           'asto.arrs.mth2arr': ('arrs.html#mth2arr', 'asto/arrs.py'),
                           'asto.arrs.mtx2arr': ('arrs.html#mtx2arr', 'asto/arrs.py'),
                           'asto.arrs.ndarray': ('arrs.html#ndarray', 'asto/arrs.py'),
                           'asto.arrs.numpy': ('arrs.html#numpy', 'asto/arrs.py'),
                           'asto.arrs.obj2arr': ('arrs.html#obj2arr', 'asto/arrs.py'),
                           'asto.arrs.spr2arr': ('arrs.html#spr2arr', 'asto/arrs.py'),
                           'asto.arrs.toarray': ('arrs.html#toarray', 'asto/arrs.py'),
                           'asto.arrs.todense': ('arrs.html#todense', 'asto/arrs.py'),
                           'asto.arrs.tolist': ('arrs.html#tolist', 'asto/arrs.py'),
                           'asto.arrs.trc2arr': ('arrs.html#trc2arr', 'asto/arrs.py'),
                           'asto.arrs.try2arr': ('arrs.html#try2arr', 'asto/arrs.py')},
            'asto.atyp': {},
            'asto.bend': { 'asto.bend.backend': ('bend.html#backend', 'asto/bend.py'),
                           'asto.bend.getbackend': ('bend.html#getbackend', 'asto/bend.py'),
                           'asto.bend.hasbackend': ('bend.html#hasbackend', 'asto/bend.py'),
                           'asto.bend.hascuda': ('bend.html#hascuda', 'asto/bend.py'),
                           'asto.bend.hasmps': ('bend.html#hasmps', 'asto/bend.py'),
                           'asto.bend.isavail': ('bend.html#isavail', 'asto/bend.py'),
                           'asto.bend.isbuilt': ('bend.html#isbuilt', 'asto/bend.py')},
            'asto.cast': { 'asto.cast.cast': ('cast.html#cast', 'asto/cast.py'),
                           'asto.cast.caster': ('cast.html#caster', 'asto/cast.py'),
                           'asto.cast.trycast': ('cast.html#trycast', 'asto/cast.py')},
            'asto.chns': { 'asto.chns._flip_channels_format': ('chns.html#_flip_channels_format', 'asto/chns.py'),
                           'asto.chns._guess_channels_index': ('chns.html#_guess_channels_index', 'asto/chns.py'),
                           'asto.chns._match_channels_format_string': ('chns.html#_match_channels_format_string', 'asto/chns.py'),
                           'asto.chns._match_channels_format_tuple': ('chns.html#_match_channels_format_tuple', 'asto/chns.py'),
                           'asto.chns.as_channels_format': ('chns.html#as_channels_format', 'asto/chns.py'),
                           'asto.chns.ascfmt': ('chns.html#ascfmt', 'asto/chns.py'),
                           'asto.chns.aschannels': ('chns.html#aschannels', 'asto/chns.py'),
                           'asto.chns.aschnfmt': ('chns.html#aschnfmt', 'asto/chns.py'),
                           'asto.chns.aschns': ('chns.html#aschns', 'asto/chns.py'),
                           'asto.chns.channels_resize': ('chns.html#channels_resize', 'asto/chns.py'),
                           'asto.chns.format_channels': ('chns.html#format_channels', 'asto/chns.py'),
                           'asto.chns.match_channels_format': ('chns.html#match_channels_format', 'asto/chns.py'),
                           'asto.chns.to_channels_format': ('chns.html#to_channels_format', 'asto/chns.py')},
            'asto.cons': {},
            'asto.dict': {},
            'asto.nums': { 'asto.nums.ascomplex': ('nums.html#ascomplex', 'asto/nums.py'),
                           'asto.nums.asfloat': ('nums.html#asfloat', 'asto/nums.py'),
                           'asto.nums.asint': ('nums.html#asint', 'asto/nums.py'),
                           'asto.nums.asnum': ('nums.html#asnum', 'asto/nums.py')},
            'asto.path': { 'asto.path.aspath': ('path.html#aspath', 'asto/path.py'),
                           'asto.path.path2str': ('path.html#path2str', 'asto/path.py'),
                           'asto.path.pathchain': ('path.html#pathchain', 'asto/path.py'),
                           'asto.path.try2path': ('path.html#try2path', 'asto/path.py')},
            'asto.sets': { 'asto.sets.any2set': ('sets.html#any2set', 'asto/sets.py'),
                           'asto.sets.asset': ('sets.html#asset', 'asto/sets.py'),
                           'asto.sets.setchain': ('sets.html#setchain', 'asto/sets.py'),
                           'asto.sets.try2set': ('sets.html#try2set', 'asto/sets.py')},
            'asto.slcs': { 'asto.slcs.asslc': ('slcs.html#asslc', 'asto/slcs.py'),
                           'asto.slcs.asslcfmt': ('slcs.html#asslcfmt', 'asto/slcs.py'),
                           'asto.slcs.asslcs': ('slcs.html#asslcs', 'asto/slcs.py'),
                           'asto.slcs.asslice': ('slcs.html#asslice', 'asto/slcs.py'),
                           'asto.slcs.flip_slice_format': ('slcs.html#flip_slice_format', 'asto/slcs.py'),
                           'asto.slcs.format_slice': ('slcs.html#format_slice', 'asto/slcs.py'),
                           'asto.slcs.guess_slice_format': ('slcs.html#guess_slice_format', 'asto/slcs.py'),
                           'asto.slcs.make_slices': ('slcs.html#make_slices', 'asto/slcs.py'),
                           'asto.slcs.makeslc': ('slcs.html#makeslc', 'asto/slcs.py'),
                           'asto.slcs.makeslcs': ('slcs.html#makeslcs', 'asto/slcs.py'),
                           'asto.slcs.makeslice': ('slcs.html#makeslice', 'asto/slcs.py'),
                           'asto.slcs.match_slice_format': ('slcs.html#match_slice_format', 'asto/slcs.py'),
                           'asto.slcs.safeslice': ('slcs.html#safeslice', 'asto/slcs.py'),
                           'asto.slcs.sdim': ('slcs.html#sdim', 'asto/slcs.py'),
                           'asto.slcs.slcchain': ('slcs.html#slcchain', 'asto/slcs.py'),
                           'asto.slcs.slice2list': ('slcs.html#slice2list', 'asto/slcs.py'),
                           'asto.slcs.slice_dimensions': ('slcs.html#slice_dimensions', 'asto/slcs.py'),
                           'asto.slcs.slice_dims': ('slcs.html#slice_dims', 'asto/slcs.py'),
                           'asto.slcs.tocoordformat': ('slcs.html#tocoordformat', 'asto/slcs.py'),
                           'asto.slcs.tosliceformat': ('slcs.html#tosliceformat', 'asto/slcs.py'),
                           'asto.slcs.try2slc': ('slcs.html#try2slc', 'asto/slcs.py'),
                           'asto.slcs.trysfmt': ('slcs.html#trysfmt', 'asto/slcs.py')},
            'asto.tens': { 'asto.tens._dev_in_sig': ('tens.html#_dev_in_sig', 'asto/tens.py'),
                           'asto.tens.asdev': ('tens.html#asdev', 'asto/tens.py'),
                           'asto.tens.asnans': ('tens.html#asnans', 'asto/tens.py'),
                           'asto.tens.assize': ('tens.html#assize', 'asto/tens.py'),
                           'asto.tens.asten': ('tens.html#asten', 'asto/tens.py'),
                           'asto.tens.devchain': ('tens.html#devchain', 'asto/tens.py'),
                           'asto.tens.dtypechain': ('tens.html#dtypechain', 'asto/tens.py'),
                           'asto.tens.gettype': ('tens.html#gettype', 'asto/tens.py'),
                           'asto.tens.mpsdtype': ('tens.html#mpsdtype', 'asto/tens.py'),
                           'asto.tens.nanslike': ('tens.html#nanslike', 'asto/tens.py'),
                           'asto.tens.ptdevice': ('tens.html#ptdevice', 'asto/tens.py'),
                           'asto.tens.pttensor': ('tens.html#pttensor', 'asto/tens.py'),
                           'asto.tens.resize': ('tens.html#resize', 'asto/tens.py'),
                           'asto.tens.resize_tensor': ('tens.html#resize_tensor', 'asto/tens.py'),
                           'asto.tens.ten2cpu': ('tens.html#ten2cpu', 'asto/tens.py'),
                           'asto.tens.ten2cuda': ('tens.html#ten2cuda', 'asto/tens.py'),
                           'asto.tens.ten2mps': ('tens.html#ten2mps', 'asto/tens.py'),
                           'asto.tens.tenchain': ('tens.html#tenchain', 'asto/tens.py'),
                           'asto.tens.to': ('tens.html#to', 'asto/tens.py'),
                           'asto.tens.torchnans': ('tens.html#torchnans', 'asto/tens.py'),
                           'asto.tens.torchtype': ('tens.html#torchtype', 'asto/tens.py'),
                           'asto.tens.totorch': ('tens.html#totorch', 'asto/tens.py'),
                           'asto.tens.try2dev': ('tens.html#try2dev', 'asto/tens.py'),
                           'asto.tens.try2dtype': ('tens.html#try2dtype', 'asto/tens.py'),
                           'asto.tens.try2move': ('tens.html#try2move', 'asto/tens.py'),
                           'asto.tens.try2ten': ('tens.html#try2ten', 'asto/tens.py')},
            'asto.time': { 'asto.time.day2hours': ('time.html#day2hours', 'asto/time.py'),
                           'asto.time.day2hrs': ('time.html#day2hrs', 'asto/time.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_cast.ipynb.

# %% auto 0
__all__ = ['cast', 'caster', 'trycast']

# %% ../nbs/02_cast.ipynb 6
from functools import wraps

# %% ../nbs/02_cast.ipynb 8
from typing import Any, Union, Optional, Callable, TypeGuard

# %% ../nbs/02_cast.ipynb 10
#| export


# %% ../nbs/02_cast.ipynb 12
from chck import isnone, notnone

# %% ../nbs/02_cast.ipynb 14
from .atyp import O, P, T

# %% ../nbs/02_cast.ipynb 16
def cast(x, dtype: T, retnone: bool = False) -> Union[T, Any, None]:
    '''Tries to cast `x` to `dtype`. If `retnone` is `True`, returns `None` if the cast fails, otherwise returns `x`.'''
    try: return dtype(x)
    except: return None if retnone else x


def caster(dtype: T, retnone: bool = False):
    def decorator(func):
        @wraps(func)
        def wrapper(x, retnone: bool = retnone) -> Union[T, Any, None]:
            return cast(x, dtype, retnone)
        return wrapper
    return decorator

# %% ../nbs/02_cast.ipynb 18
def trycast(
    x: O, *fns: Callable[P, T], 
    retnone: bool = False, 
    guardfn: Optional[Callable[P, TypeGuard[T]]] = None,
) -> Union[T, Any, None]:
    '''Attempts to cast `x` using provided functions `fns`, returning `None` 
    or the original `x` if unsuccessful depending on `retnone`.
    
    Parameters
    ----------
    x : O
        The object to attempt to cast.
        
    *fns : Callable[P, T]
        Functions to attempt casting with.
        
    retnone : bool
        Whether to return `None` if casting fails.
        
    guardfn : Optional[Callable[P, TypeGuard[T]]]
        An optional guard function to check if `x` is already of the desired type.
    
    Returns
    -------
    Union[T, Any, None]
        The casted object, original object, or `None` if casting fails and `retnone` is True.
    '''
    
    # Short-circuit if `x` is already of the desired type
    if callable(guardfn) and guardfn(x): return x
    for fn in fns:
        try: 
            ret = fn(x)
            if notnone(ret): return ret
            if not retnone and isnone(ret): continue
            break
        except: ...
    return None if retnone else x



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_sets.ipynb.

# %% auto 0
__all__ = ['setchain', 'try2set', 'any2set', 'asset']

# %% ../nbs/05_sets.ipynb 6
#| export


# %% ../nbs/05_sets.ipynb 8
#| export


# %% ../nbs/05_sets.ipynb 10
#| export


# %% ../nbs/05_sets.ipynb 12
from nlit import __HASH__
from sigr import applyfns
from chck import isiter, islist, istuple, isset, isiter, isiternotstr

# %% ../nbs/05_sets.ipynb 14
from .atyp import P

# %% ../nbs/05_sets.ipynb 17
def setchain(x, funcs, *args: P.args, **kwargs: P.kwargs): 
    '''Applies a sequence of functions (`funcs`) to an initial value `x`, 
    optionally stopping when the typeguard function `isset` returns `True`.
    
    See Also
    --------
    sigr.applyfns
    ''' 
    return applyfns(x, funcs, *args, check=isset, **kwargs)

# %% ../nbs/05_sets.ipynb 19
def try2set(a) -> set:
    '''Attempts to convert `a` to a set. 
    If `a` is already a set, returns it directly; otherwise, tries to create a set from `a`.
    
    Parameters
    ----------
    a : Any
        The input object to convert to a set.
        
    Returns
    -------
    set
        A set created from `a` if possible; otherwise, returns `a` itself if it cannot be converted.
    '''
    if isset(a): return a
    try: return set([a]) if not isiternotstr(a) else set(a)
    except: return a

# %% ../nbs/05_sets.ipynb 21
def any2set(obj) -> set:
    '''Attempts to convert any object to a set, handling strings and non-iterable 
    objects by creating a set of length 1 with the object.
    
    Parameters
    ----------
    obj : Any
        The object to convert to a set.
        
    Returns
    -------
    set
        The resulting set.  
        
    Notes
    -----
    - If the object is a string, it will be converted to a set of length 1.
    - If `obj` is a string or not iterable, the set will contain `obj` as its only item.
    - If `obj` is iterable, it converts the iterable to a set, unless `obj` is already a set.
    '''
    if isinstance(obj, str): obj = set((obj, ))
    elif not isiter(obj): obj = set((obj, )) # not a string and not iterable

    elif any(c(obj) for c in (isiter, islist, istuple)) and not isset(obj):    
        obj = set(obj)
    elif hasattr(obj, __HASH__): obj = {obj, }
    else:
        try: obj = set(obj)
        except TypeError: obj = set()
    return obj if isset(obj) else set()

# %% ../nbs/05_sets.ipynb 23
def asset(a, *args, **kwargs) -> set:
    '''Converts an input to a set.
    
    Parameters
    ----------
    a : Any
        The input to convert.
    
    *args : Any
        Additional arguments for conversion and adjustment functions.
        
    **kwargs : Any
        Additional keyword arguments for conversion and adjustment functions.
        
    Returns
    -------
    set
        The resulting set.
    '''
    if isset(a): return a
    chain = (any2set, try2set, )
    return setchain(a, chain, *args, **kwargs)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_time.ipynb.

# %% auto 0
__all__ = ['day2hrs', 'day2hours']

# %% ../nbs/04_time.ipynb 6
from functools import wraps
from datetime import date, datetime

# %% ../nbs/04_time.ipynb 8
#| export


# %% ../nbs/04_time.ipynb 10
#| export


# %% ../nbs/04_time.ipynb 12
from nlit import __DOC__, __NAME__, __MODULE__, __ANNOTATIONS__

# %% ../nbs/04_time.ipynb 14
#| export


# %% ../nbs/04_time.ipynb 16
def day2hrs(day: int) -> int:
    '''Converts a quantity in days to the equivalent number of hours.
    
    Parameters
    ----------
    day : int
        The number of days to convert.
        
    Returns
    -------
    int
        The equivalent number of hours.
        
    See Also
    --------
    day2hours : Alias of `day2hrs`.
    '''
    return int(day * 24)

@wraps(day2hrs, assigned=(__DOC__, __NAME__, __MODULE__, __ANNOTATIONS__))
def day2hours(day: int) -> int:
    '''Alias of `day2hrs`'''
    return day2hrs(day)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/60_bend.ipynb.

# %% auto 0
__all__ = ['isavail', 'isbuilt', 'getbackend', 'hasbackend', 'backend', 'hasmps', 'hascuda']

# %% ../nbs/60_bend.ipynb 6
from functools import wraps

# %% ../nbs/60_bend.ipynb 8
from types import ModuleType
from typing import Callable

# %% ../nbs/60_bend.ipynb 11
#| export

# %% ../nbs/60_bend.ipynb 13
#| export


# %% ../nbs/60_bend.ipynb 15
try: import torch
except ImportError: ...

# %% ../nbs/60_bend.ipynb 17
#| export


# %% ../nbs/60_bend.ipynb 19
from nlit import IS_AVAILABLE, IS_BUILT, MPS, CUDA
from nlit import __DOC__, __NAME__, __MODULE__, __QUALNAME__

# %% ../nbs/60_bend.ipynb 21
from .atyp import P

# %% ../nbs/60_bend.ipynb 24
def isavail(backend: ModuleType) -> bool:
    '''Check if the given backend is available.

    Parameters
    ----------
    backend : ModuleType
        The backend module to check for availability.

    Returns
    -------
    bool
        True if the backend is available, False otherwise.
    '''
    is_avail = getattr(backend, IS_AVAILABLE, lambda : False)
    return is_avail()

def isbuilt(backend: ModuleType) -> bool:
    '''Check if the given backend is built.

    Parameters
    ----------
    backend : ModuleType
        The backend module to check for build status.

    Returns
    -------
    bool
        True if the backend is built, False otherwise.
    '''
    is_built = getattr(backend, IS_BUILT, lambda : False)
    return is_built()

def getbackend(name: str) -> ModuleType | None:
    '''Retrieve a backend module by name.

    Parameters
    ----------
    name : str
        The name of the backend to retrieve.

    Returns
    -------
    ModuleType | None
        The backend module if found, None otherwise.
    '''
    if isinstance(name, ModuleType): return name
    try: return getattr(torch.backends, name)
    except AttributeError: return None

def hasbackend(backend: str) -> bool:
    '''Check if a backend is both available and built.

    Parameters
    ----------
    backend : str
        The name of the backend to check.

    Returns
    -------
    bool
        True if the backend is available and built, False otherwise.
    '''
    be = getbackend(backend)
    return isbuilt(be) and isavail(be)

# %% ../nbs/60_bend.ipynb 26
def backend(name: str) -> Callable[[Callable], Callable[[], bool]]:
    '''Decorator to check if a specified backend is built and available.

    Parameters
    ----------
    name : str
        The name of the backend to check.

    Returns
    -------
    Callable[[Callable], bool]
        A decorator function that when applied to a function, returns the result of `hasbackend(name)`.
    '''
    def decorator(func: Callable) -> Callable[[], bool]:
        func.__doc__ = f'''Checks whether or not pytorch has {name} availble and was built with {name} in mind.'''
        @wraps(func, assigned=(__NAME__, __MODULE__, __QUALNAME__, __DOC__,))
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> bool:
            return hasbackend(name)
        return wrapper
    return decorator

# %% ../nbs/60_bend.ipynb 28
@backend(MPS)
def hasmps() -> bool: 
    '''Checks whether or not pytorch has mps availble and was built with mps in mind.'''
    
@backend(CUDA)
def hascuda() -> bool:
    '''Checks whether or not pytorch has cuda availble and was built with cuda in mind.'''

# %% ../nbs/60_bend.ipynb 30
#| export



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/30_slcs.ipynb.

# %% auto 0
__all__ = ['safeslice', 'sdim', 'slice_dimensions', 'slice_dims', 'match_slice_format', 'guess_slice_format', 'flip_slice_format',
           'slcchain', 'slice2list', 'format_slice', 'asslcfmt', 'tosliceformat', 'tocoordformat', 'trysfmt', 'try2slc',
           'asslc', 'asslice', 'makeslc', 'makeslice', 'makeslcs', 'make_slices', 'asslcs']

# %% ../nbs/30_slcs.ipynb 6
from functools import wraps, partial

# %% ../nbs/30_slcs.ipynb 8
#| export


# %% ../nbs/30_slcs.ipynb 10
#| export


# %% ../nbs/30_slcs.ipynb 12
from nlit import (COORD, SLICE,)
from quac import ints, intq, slicespec, slicespecs
from chck import isnone, notnone, isslice, isiter
from sigr import applyfns

# %% ../nbs/30_slcs.ipynb 14
from .cons import WRAPS_ASSIGN_ANNDOCS
from .atyp import P, SliceFormats, SliceSpec, SliceSpecs

# %% ../nbs/30_slcs.ipynb 17
def safeslice(s: slice) -> slice:
    '''Corrects a slice object, replacing None values with defaults to prevent errors.
    
    Returns a slice with default values instead None values from the input slice.
    e.g. slice(IntQ, IntQ, IntQ) -> slice(0, 1, 1)
    
    Parameters
    ----------
    s : slice
        The slice object to correct.
        
    Returns
    -------
    slice
        A corrected slice object with no None values.
    '''
    if not isslice(s): return s
    start = s.start if notnone(s.start) else 0
    stop  = s.stop  if notnone(s.stop)  else min(0, start, max(1, start))
    step  = s.step  if notnone(s.step)  else 1
    return slice(start, stop, step)

# %% ../nbs/30_slcs.ipynb 19
def sdim(*slcs: SliceSpecs, axes: ints | None = None, ndim: intq = None) -> int:
     '''
     Calculate the number of dimensions to which the slices apply.

     Parameters
     ----------
     slcs : slicespecs | tuple[tuple[int, int, int], ...] | tuple[tuple[int, ...], ...]
          The slices for which the dimensionality is to be determined.

     axes : tuple[int, ...], optional
          The axes that the slices apply to, defaults to None.

     ndim : int, optional
          The number of dimensions of the target tensor, defaults to None.

     Returns
     -------
     int
          The number of dimensions to which the slices apply.

     Notes
     -----
     slice : SliceFormats
          The default representations of a python slice, i.e. 
          `(start: int, stop: int, step: int)`
          
     coord : SliceFormats
          The coordinate repesentation of a python slice i.e. 
          `(x0, y0, ...), (x1, y1, ...), ...`
          
     __nones_ok : bool
          A sometimes optional parameter for slice utility functions.
          If `True`, the result is not corrected to prevent None values.
          i.e. `slice(0, 10, None)` would become `slice(0, 10, 1)`
          
                    
     Definitions
     -----------
     :slice: `slice`
          The standard representation for a python slice, e.g.
          `(start: int, stop: int, step: int)` or `slice(0, 10, 2)`


     :svals: `tuple[int | None, int, int | None]`
          The integer representation for a python slice, e.g.
          `(start: int | None, stop: int, step: int | None)` or `slice(0, 10, None)`
          
     :slices: `tuple[slice | svals, ...] `
          An iterable of like representation for a python slice, e.g.
          `slice(0, 10, 2)`, `tuple(0, 10, 2)`, `tuple(0, 10, None)`, etc

     :coord: `tuple[int, ...]`
          The coordinate representation for a python slice, e.g.
          `(x0, y0, ...), (x1, y1, ...), ...`
          
     :coords: `tuple[coord, ...]`
          An iterable of like coordinate representation for a python slice, e.g.
          `(x0, y0, ...), (x1, y1, ...), ...`
    
     Examples
     -------- 
     >>> sdim((0, 10), (2, 10))
     2

     >>> slices = (0, 5), (2, 10) # slice format: (x0, x1), (y0, y1)
     ... tocoordformat(slices) # convert to coordinate format
     ((0, 2), (5, 10)) 

     >>> coords = (0, 2), (5, 10) # coord format: (x0, y0), (x1, y1)
     ... tosliceformat(coords) # convert to slice format
     ((0, 5), (2, 10))

     See Also
     --------
     slice_dims : Alias for sdim
     slice_dimensions : Alias for sdim
     asto.slcs.tocoordformat: convert slices in slice format to coordinate format
     asto.slcs.tosliceformat: convert slices in coordinate format to slice format
     '''
     if notnone(ndim): 
          return ndim
     
     if isnone(axes): 
          return len(slcs)
     
     if isiter(axes): 
          return max(max(axes) + 1, len(axes))
     
     return axes

@wraps(sdim, assigned=WRAPS_ASSIGN_ANNDOCS)
def slice_dimensions(*slcs: slicespecs, axes: ints | None = None, ndim: intq = None) -> int:
     '''Alias for `sdim`'''
     return sdim(*slcs, axes=axes, ndim=ndim)

@wraps(sdim, assigned=WRAPS_ASSIGN_ANNDOCS)
def slice_dims(*slcs: slicespecs, axes: ints | None = None, ndim: intq = None) -> int:
     '''Alias for `sdim`'''
     return sdim(*slcs, axes=axes, ndim=ndim)

# %% ../nbs/30_slcs.ipynb 22
def match_slice_format(fmt: str) -> SliceFormats:
    '''Matches the given format to the closest valid format.
    Parameters
    ----------
    fmt : str
        The format to match.
    
    Returns
    -------
    fmt : {`'coord'`, `'slice'`}
        The matched slice format.
        
    Notes
    -----
    slice : SliceFormats
        The default representations of a python slice, i.e. 
        `(start: int, stop: int, step: int)`
        
    coord : SliceFormats
        The coordinate repesentation of a python slice i.e. 
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    __nones_ok : bool
        A sometimes optional parameter for slice utility functions.
        If `True`, the result is not corrected to prevent None values.
        i.e. `slice(0, 10, None)` would become `slice(0, 10, 1)`
        
                
    Definitions
    -----------
    :slice: `slice`
        The standard representation for a python slice, e.g.
        `(start: int, stop: int, step: int)` or `slice(0, 10, 2)`


    :svals: `tuple[int | None, int, int | None]`
        The integer representation for a python slice, e.g.
        `(start: int | None, stop: int, step: int | None)` or `slice(0, 10, None)`
        
    :slices: `tuple[slice | svals, ...] `
        An iterable of like representation for a python slice, e.g.
        `slice(0, 10, 2)`, `tuple(0, 10, 2)`, `tuple(0, 10, None)`, etc

    :coord: `tuple[int, ...]`
        The coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    :coords: `tuple[coord, ...]`
        An iterable of like coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`

    Examples
    -------- 
    >>> sdim((0, 10), (2, 10))
    2

    >>> slices = (0, 5), (2, 10) # slice format: (x0, x1), (y0, y1)
    ... tocoordformat(slices) # convert to coordinate format
    ((0, 2), (5, 10)) 

    >>> coords = (0, 2), (5, 10) # coord format: (x0, y0), (x1, y1)
    ... tosliceformat(coords) # convert to slice format
    ((0, 5), (2, 10))

    See Also
    --------
    slice_dims : Alias for sdim
    slice_dimensions : Alias for sdim
    asto.slcs.tocoordformat: convert slices in slice format to coordinate format
    asto.slcs.tosliceformat: convert slices in coordinate format to slice format
    '''
    match fmt:
        case 'coord': return COORD
        case 'slice': return SLICE
        case _: return SLICE

# %% ../nbs/30_slcs.ipynb 24
def guess_slice_format(*slcs: SliceSpecs) -> SliceFormats:
    '''Guesses the slice format based on the given slices.

    Parameters
    ----------
    slcs : slicespecs | tuple[tuple[int, int, int], ...] | tuple[tuple[int, ...], ...]
        The slices to evaluate.

    Returns
    -------
    fmt : {`'coord'`, `'slice'`}
        The guessed slice format.
        
    Notes
    -----
    slice : SliceFormats
        The default representations of a python slice, i.e. 
        `(start: int, stop: int, step: int)`
        
    coord : SliceFormats
        The coordinate repesentation of a python slice i.e. 
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    __nones_ok : bool
        A sometimes optional parameter for slice utility functions.
        If `True`, the result is not corrected to prevent None values.
        i.e. `slice(0, 10, None)` would become `slice(0, 10, 1)`
        
                
    Definitions
    -----------
    :slice: `slice`
        The standard representation for a python slice, e.g.
        `(start: int, stop: int, step: int)` or `slice(0, 10, 2)`


    :svals: `tuple[int | None, int, int | None]`
        The integer representation for a python slice, e.g.
        `(start: int | None, stop: int, step: int | None)` or `slice(0, 10, None)`
        
    :slices: `tuple[slice | svals, ...] `
        An iterable of like representation for a python slice, e.g.
        `slice(0, 10, 2)`, `tuple(0, 10, 2)`, `tuple(0, 10, None)`, etc

    :coord: `tuple[int, ...]`
        The coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    :coords: `tuple[coord, ...]`
        An iterable of like coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`

    Examples
    -------- 
    >>> guess_slice_format((0, 10), (0, 10))
    'slice'

    >>> slices = (0, 5), (2, 10) # slice format: (x0, x1), (y0, y1)
    ... tocoordformat(slices) # convert to coordinate format
    ((0, 2), (5, 10)) 

    >>> coords = (0, 2), (5, 10) # coord format: (x0, y0), (x1, y1)
    ... tosliceformat(coords) # convert to slice format
    ((0, 5), (2, 10))

    See Also
    --------
    asto.slcs.tocoordformat: convert slices in slice format to coordinate format
    asto.slcs.tosliceformat: convert slices in coordinate format to slice format
    '''
    if any(isslice(slc) for slc in slcs): return SLICE
    if any([len(slc) > 3 for slc in slcs]): return COORD
    return SLICE

# %% ../nbs/30_slcs.ipynb 26
def flip_slice_format(*slcs: SliceSpecs) -> SliceSpecs:
    '''
    Flips the format of the given slices between 'coord' and 'slice'.

    Parameters
    ----------
    slcs : slicespecs | tuple[tuple[int, int, int], ...] | tuple[tuple[int, ...], ...]
        The slices to flip. Can be in 'coord' or 'slice' format.

    Returns
    -------
    slices: slicespecs | tuple[tuple[int, int, int], ...] | tuple[tuple[int, ...], ...]
        The slices flipped to the opposite format.

    Notes
    -----
    slice : SliceFormats
        The default representations of a python slice, i.e. 
        `(start: int, stop: int, step: int)`
        
    coord : SliceFormats
        The coordinate repesentation of a python slice i.e. 
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    __nones_ok : bool
        A sometimes optional parameter for slice utility functions.
        If `True`, the result is not corrected to prevent None values.
        i.e. `slice(0, 10, None)` would become `slice(0, 10, 1)`
        
                
    Definitions
    -----------
    :slice: `slice`
        The standard representation for a python slice, e.g.
        `(start: int, stop: int, step: int)` or `slice(0, 10, 2)`


    :svals: `tuple[int | None, int, int | None]`
        The integer representation for a python slice, e.g.
        `(start: int | None, stop: int, step: int | None)` or `slice(0, 10, None)`
        
    :slices: `tuple[slice | svals, ...] `
        An iterable of like representation for a python slice, e.g.
        `slice(0, 10, 2)`, `tuple(0, 10, 2)`, `tuple(0, 10, None)`, etc

    :coord: `tuple[int, ...]`
        The coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    :coords: `tuple[coord, ...]`
        An iterable of like coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`

    Examples
    --------    
    >>> flip_slice_format((0, 10), (0, 10))
    ((0, 0), (10, 10)) # coordinate format

    >>> slices = (0, 5), (2, 10) # slice format: (x0, x1), (y0, y1)
    ... tocoordformat(slices) # convert to coordinate format
    ((0, 2), (5, 10)) 

    >>> coords = (0, 2), (5, 10) # coord format: (x0, y0), (x1, y1)
    ... tosliceformat(coords) # convert to slice format
    ((0, 5), (2, 10))

    See Also
    --------
    asto.slcs.tocoordformat: convert slices in slice format to coordinate format
    asto.slcs.tosliceformat: convert slices in coordinate format to slice format
    '''
    return tuple(zip(*slcs))

# %% ../nbs/30_slcs.ipynb 28
def slcchain(x, funcs, *args: P.args, **kwargs: P.kwargs): 
    '''Applies a sequence of functions (`funcs`) to an initial value `x`, 
    optionally stopping when the typeguard function `isslice` returns `True`.
    
    See Also
    --------
    sigr.applyfns
    ''' 
    return applyfns(x, funcs, *args, check=isslice, **kwargs)

# %% ../nbs/30_slcs.ipynb 31
def slice2list(s: slice) -> tuple[intq, int, intq]:
    '''Converts a slice object to a list of integers it represents.
    
    Returns a list of integers from a slice.
    
    Parameters
    ----------
    s : slice
        The slice object to convert.
        
    Returns
    -------
    list
        A list of integers represented by the slice.
    '''
    s = safeslice(s)
    try:
        return list(range(s.start, s.stop, s.step))
    except: 
        return list()

# %% ../nbs/30_slcs.ipynb 33
def format_slice(*slcs: SliceSpecs, fmt: SliceFormats = SLICE, cur: SliceFormats | None = None) -> SliceSpecs:
    '''
    Format the given slices as either coordinates ('coord') or 'slice'.

    Parameters
    ----------
    slcs : slicespecs | tuple[tuple[int, int, int], ...] | tuple[tuple[int, ...], ...]
        The slices to flip. Can be in 'coord' or 'slice' format.

    Returns
    -------
    slices: slicespecs | tuple[tuple[int, int, int], ...] | tuple[tuple[int, ...], ...]
        The slices flipped to the opposite format.
    
    Notes
    -----
    slice : SliceFormats
        The default representations of a python slice, i.e. 
        `(start: int, stop: int, step: int)`
        
    coord : SliceFormats
        The coordinate repesentation of a python slice i.e. 
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    __nones_ok : bool
        A sometimes optional parameter for slice utility functions.
        If `True`, the result is not corrected to prevent None values.
        i.e. `slice(0, 10, None)` would become `slice(0, 10, 1)`
                
    Definitions
    -----------
    :slice: `slice`
        The standard representation for a python slice, e.g.
        `(start: int, stop: int, step: int)` or `slice(0, 10, 2)`


    :svals: `tuple[int | None, int, int | None]`
        The integer representation for a python slice, e.g.
        `(start: int | None, stop: int, step: int | None)` or `slice(0, 10, None)`
        
    :slices: `tuple[slice | svals, ...] `
        An iterable of like representation for a python slice, e.g.
        `slice(0, 10, 2)`, `tuple(0, 10, 2)`, `tuple(0, 10, None)`, etc

    :coord: `tuple[int, ...]`
        The coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    :coords: `tuple[coord, ...]`
        An iterable of like coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`

    Examples
    --------
    >>> slices = (0, 5), (2, 10) # slice format: (x0, x1), (y0, y1)
    ... tocoordformat(slices) # convert to coordinate format
    ((0, 2), (5, 10)) 

    >>> coords = (0, 2), (5, 10) # coord format: (x0, y0), (x1, y1)
    ... tosliceformat(coords) # convert to slice format
    ((0, 5), (2, 10))

    See Also
    --------
    asto.slcs.tocoordformat: convert slices in slice format to coordinate format
    asto.slcs.tosliceformat: convert slices in coordinate format to slice format
    '''
    fmt = match_slice_format(fmt)
    if isnone(cur): 
        cur = guess_slice_format(*slcs)
        
    if cur == fmt: 
        return slcs
    
    else: 
        return flip_slice_format(*slcs)
    
@wraps(format_slice, assigned=WRAPS_ASSIGN_ANNDOCS)
def asslcfmt(*slcs, fmt: SliceFormats = SLICE, cur: SliceFormats | None = None) -> slicespecs:
    return format_slice(*slcs, fmt = fmt, cur = cur)

@wraps(format_slice, assigned=WRAPS_ASSIGN_ANNDOCS)
def tosliceformat(*slcs, cur: SliceFormats | None = None) -> slicespecs:
    formatter = partial(format_slice, fmt=SLICE)
    return formatter(*slcs, cur = cur)

@wraps(format_slice, assigned=WRAPS_ASSIGN_ANNDOCS)
def tocoordformat(*slcs, cur: SliceFormats | None = None) -> slicespecs:
    formatter = partial(format_slice, fmt=COORD)
    return formatter(*slcs, cur = cur)

# %% ../nbs/30_slcs.ipynb 35
def trysfmt(a: slice, fmt: SliceFormats | None = None, cur: SliceFormats | None = None) -> SliceSpec:
    '''Attempts to format `a` to the desired format.
    
    Parameters
    ----------
    a : Any
        The input object to convert.
        
    Returns
    -------
    slice
        A slice object if conversion is successful, otherwise `a`.
        
    Notes
    -----
    slice : SliceFormats
        The default representations of a python slice, i.e. 
        `(start: int, stop: int, step: int)`
        
    coord : SliceFormats
        The coordinate repesentation of a python slice i.e. 
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    __nones_ok : bool
        A sometimes optional parameter for slice utility functions.
        If `True`, the result is not corrected to prevent None values.
        i.e. `slice(0, 10, None)` would become `slice(0, 10, 1)`
                
    Definitions
    -----------
    :slice: `slice`
        The standard representation for a python slice, e.g.
        `(start: int, stop: int, step: int)` or `slice(0, 10, 2)`


    :svals: `tuple[int | None, int, int | None]`
        The integer representation for a python slice, e.g.
        `(start: int | None, stop: int, step: int | None)` or `slice(0, 10, None)`
        
    :slices: `tuple[slice | svals, ...] `
        An iterable of like representation for a python slice, e.g.
        `slice(0, 10, 2)`, `tuple(0, 10, 2)`, `tuple(0, 10, None)`, etc

    :coord: `tuple[int, ...]`
        The coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    :coords: `tuple[coord, ...]`
        An iterable of like coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`

    Examples
    --------
    >>> slices = (0, 5), (2, 10) # slice format: (x0, x1), (y0, y1)
    ... tocoordformat(slices) # convert to coordinate format
    ((0, 2), (5, 10)) 

    >>> coords = (0, 2), (5, 10) # coord format: (x0, y0), (x1, y1)
    ... tosliceformat(coords) # convert to slice format
    ((0, 5), (2, 10))
    
    See Also
    --------
    asto.slcs.tocoordformat: convert slices in slice format to coordinate format
    asto.slcs.tosliceformat: convert slices in coordinate format to slice format
    '''
    if isslice(a) and isnone(fmt): return a
    try: return format_slice(a, fmt, cur)
    except: return a

# %% ../nbs/30_slcs.ipynb 37
def try2slc(a) -> slice:
    '''Attempts to convert `a` to a slice object, falling back to `a` if unsuccessful.
    
    Parameters
    ----------
    a : Any
        The input object to convert.
        
    Returns
    -------
    slice
        A slice object if conversion is successful, otherwise `a`.
    '''
    if isslice(a): return a
    try: return slice(*list((a, )if not isiter(a) else a))    
    except: return a

# %% ../nbs/30_slcs.ipynb 39
def asslc(a, *args: P.args, **kwargs: P.kwargs) -> slice:
    '''Ensures `a` is represented as or converted to a slice object, 
    applying corrections and conversions as needed.
    
    Parameters
    ----------
    a : Any
        The input object to convert.
        
    *args : P.args
        Additional positional arguments for conversion functions.
        
    **kwargs : P.kwargs
        Additional keyword arguments for conversion functions.
        
    Other Parameters
    ----------------
    __format: SliceFormats, optional
        The format to coerce the slice into.
        
    __current: SliceFormats, optional
        The current to format the slice is in.
        
    __nones_ok : bool, default: False
        If `True`, the result is not corrected to prevent None values.
        i.e. `slice(0, 10, None)` would become `slice(0, 10, 1)`
        
    Returns
    -------
    slice
        A slice object representing `a`, corrected if necessary.
    
    Notes
    -----
    slice : SliceFormats
        The default representations of a python slice, i.e. 
        `(start: int, stop: int, step: int)`
        
    coord : SliceFormats
        The coordinate repesentation of a python slice i.e. 
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    __nones_ok : bool
        A sometimes optional parameter for slice utility functions.
        If `True`, the result is not corrected to prevent None values.
        i.e. `slice(0, 10, None)` would become `slice(0, 10, 1)`
                
    Definitions
    -----------
    :slice: `slice`
        The standard representation for a python slice, e.g.
        `(start: int, stop: int, step: int)` or `slice(0, 10, 2)`


    :svals: `tuple[int | None, int, int | None]`
        The integer representation for a python slice, e.g.
        `(start: int | None, stop: int, step: int | None)` or `slice(0, 10, None)`
        
    :slices: `tuple[slice | svals, ...] `
        An iterable of like representation for a python slice, e.g.
        `slice(0, 10, 2)`, `tuple(0, 10, 2)`, `tuple(0, 10, None)`, etc

    :coord: `tuple[int, ...]`
        The coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    :coords: `tuple[coord, ...]`
        An iterable of like coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`

    Examples
    --------
    >>> slices = (0, 5), (2, 10) # slice format: (x0, x1), (y0, y1)
    ... tocoordformat(slices) # convert to coordinate format
    ((0, 2), (5, 10)) 

    >>> coords = (0, 2), (5, 10) # coord format: (x0, y0), (x1, y1)
    ... tosliceformat(coords) # convert to slice format
    ((0, 5), (2, 10))
        
    See Also
    --------
    asslice : alias for asslc
    makeslc : alias for asslc
    makeslice : alias for asslc
    asto.slcs.tocoordformat: convert slices in slice format to coordinate format
    asto.slcs.tosliceformat: convert slices in coordinate format to slice format
    '''
    if isslice(a): return a
    chain = (safeslice, try2slc, )
    
    nok = kwargs.pop('__nones_ok', False)
    fmt = kwargs.pop('__format', None)
    cur = kwargs.pop('__current', None)
    
    slc = slcchain(a, chain, *args, **kwargs)
    
    # 
    if nok: 
        return slc
    
    slc = safeslice(slc)
    if notnone(fmt): 
        slc = trysfmt(slc, fmt, cur)
    return slc

@wraps(asslc, assigned=WRAPS_ASSIGN_ANNDOCS)
def asslice(a, *args: P.args, **kwargs: P.kwargs) -> slice:
    '''Alias for asslc'''
    kwargs.pop('__format', None)
    return asslc(a, *args, **kwargs, __format=SLICE)

@wraps(asslc, assigned=WRAPS_ASSIGN_ANNDOCS)
def makeslc(a, *args: P.args, **kwargs: P.kwargs) -> slice:
    '''Alias for asslc'''
    return asslc(a, *args, **kwargs)

@wraps(asslc, assigned=WRAPS_ASSIGN_ANNDOCS)
def makeslice(a, *args: P.args, **kwargs: P.kwargs) -> slice:
    '''Alias for asslc'''
    return asslc(a, *args, **kwargs)

# %% ../nbs/30_slcs.ipynb 41
def makeslcs(
    *slcs: SliceSpecs, 
    axes: ints | None = None, 
    ndim: intq = None, 
    fmt: SliceFormats = SLICE,
) -> list[slice]:     
    '''Convert tuples representing slices into actual Python slice objects.

    Parameters
    ----------
    slcs : slicespecs | tuple[tuple[int, int, int], ...] | tuple[tuple[int, ...], ...]
        Tuples representing the slices.
        
    axes : tuple[int, ...], optional
        The axes that the slices apply to, defaults to None.

    ndim : int, optional
        The number of dimensions of the target tensor, defaults to None.

    fmt : SliceFormats, default: 'slice'
        The format of the input slices, either 'coord' or 'slice', defaults to 'slice'.

    Returns
    -------
    list[slice]
        A list of Python slice objects (i.e. slices in slice format)
        
    Notes
    -----
    slice : SliceFormats
        The default representations of a python slice, i.e. 
        `(start: int, stop: int, step: int)`
        
    coord : SliceFormats
        The coordinate repesentation of a python slice i.e. 
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    __nones_ok : bool
        A sometimes optional parameter for slice utility functions.
        If `True`, the result is not corrected to prevent None values.
        i.e. `slice(0, 10, None)` would become `slice(0, 10, 1)`
                
    Definitions
    -----------
    :slice: `slice`
        The standard representation for a python slice, e.g.
        `(start: int, stop: int, step: int)` or `slice(0, 10, 2)`


    :svals: `tuple[int | None, int, int | None]`
        The integer representation for a python slice, e.g.
        `(start: int | None, stop: int, step: int | None)` or `slice(0, 10, None)`
        
    :slices: `tuple[slice | svals, ...] `
        An iterable of like representation for a python slice, e.g.
        `slice(0, 10, 2)`, `tuple(0, 10, 2)`, `tuple(0, 10, None)`, etc

    :coord: `tuple[int, ...]`
        The coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`
        
    :coords: `tuple[coord, ...]`
        An iterable of like coordinate representation for a python slice, e.g.
        `(x0, y0, ...), (x1, y1, ...), ...`

    Examples
    --------
    >>> makeslcs((0, 10), (0, 10))
    [slice(0, 10, None), slice(0, 10, None)]
    
    >>> slices = (0, 5), (2, 10) # slice format: (x0, x1), (y0, y1)
    ... tocoordformat(slices) # convert to coordinate format
    ((0, 2), (5, 10)) 

    >>> coords = (0, 2), (5, 10) # coord format: (x0, y0), (x1, y1)
    ... tosliceformat(coords) # convert to slice format
    ((0, 5), (2, 10))
    
    See Also
    --------
    asslcs : Alias for makeslcs
    make_slices : Alias for makeslcs
    asto.slcs.tocoordformat: convert slices in slice format to coordinate format
    asto.slcs.tosliceformat: convert slices in coordinate format to slice format
    '''
    # (x0, y0), (x1, y1) --> (x0, x1), (y0, y1)
    if isnone(fmt): 
        fmt = guess_slice_format(*slcs)
    
    if fmt == COORD: 
        slcs = flip_slice_format(*slcs)

    # Convert slcs tuples into slices
    slices = list(asslc(slc, __nones_ok=True) for slc in slcs)
    nslice = sdim(*slices, axes=axes, ndim=ndim)
    
    # Create a list of slice(None) for all dimensions
    axes = list([slice(None)] * nslice if isnone(axes) else axes)

    # Replace slices for specified axes
    slices = [slices[axes.index(i)] if i in axes else slice(None) for i in range(nslice)]
    return slices

@wraps(makeslcs, assigned=WRAPS_ASSIGN_ANNDOCS)
def make_slices(
    *slcs: slicespecs, 
    axes: ints | None = None, 
    ndim: intq = None, 
    fmt: SliceFormats = SLICE
) -> list[slice]:
    '''Alias for makeslcs'''
    return makeslcs(*slcs, axes=axes, ndim=ndim, fmt=fmt)

@wraps(makeslcs, assigned=WRAPS_ASSIGN_ANNDOCS)
def asslcs(
    *slcs: slicespecs, 
    axes: ints | None = None, 
    ndim: intq = None, 
    fmt: SliceFormats = SLICE
) -> list[slice]:
    '''Alias for makeslcs'''
    return makeslcs(*slcs, axes=axes, ndim=ndim, fmt=fmt)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/32_arng.ipynb.

# %% auto 0
__all__ = ['isarng', 'arngchain', 'try2arng', 'arr2arng', 'df2arng', 'asarng', 'arange']

# %% ../nbs/32_arng.ipynb 6
from functools import wraps

# %% ../nbs/32_arng.ipynb 8
from typing import TypeGuard

# %% ../nbs/32_arng.ipynb 10
try: import numpy as np
except ModuleNotFoundError: np = None

# %% ../nbs/32_arng.ipynb 12
from nlit import __DOC__, __NAME__, __MODULE__, __ANNOTATIONS__
from quac import nparray, indexlike, dataframe
from chck import isarr
from sigr import applyfns

# %% ../nbs/32_arng.ipynb 14
from .atyp import P
from .arrs import asnp

# %% ../nbs/32_arng.ipynb 17
def _safe_idx(x): 
    '''try: x.index except: x
    
    Attempts to return the `.index` method of `x`, falling back to `x` itself if unsuccessful.
    
    Parameters
    ----------
    x : Any
        The input object from which to attempt retrieving the `.index` method.
        
    Returns
    -------
    Any
        The `.index` method of `x` if it exists, otherwise `x` itself.
    '''
    try: return x.index
    except: return x

def _safe_len(x):
    '''try: len(x) except: 1
    
    Attempts to return the length of `x`, defaulting to 1 if unsuccessful.
    
    Parameters
    ----------
    x : Any
        The input object from which to attempt measuring the length.
        
    Returns
    -------
    int
        The length of `x` if it is measurable, otherwise 1.
    '''
    try: return len(x)
    except: return 1



# %% ../nbs/32_arng.ipynb 19
def isarng(a) -> TypeGuard[nparray]:
    '''Checks if `a` is a numpy array with elements forming an arithmetic sequence starting at 0 with a step of 1.
    
    Parameters
    ----------
    a : Any
        The input object to check.
        
    Returns
    -------
    bool
        True if `a` is a numpy array with elements forming an arithmetic sequence, otherwise False.
    '''
    return isarr(a) and np.all(np.ediff1d(a) == 1)

# %% ../nbs/32_arng.ipynb 21
def arngchain(x, funcs, *args: P.args, **kwargs: P.kwargs): 
    '''Applies a sequence of functions (`funcs`) to an initial value `x`, 
    optionally stopping when the typeguard function `isarng` returns `True`.
    
    See Also
    --------
    sigr.applyfns
    ''' 
    return applyfns(x, funcs, *args, check=isarng, **kwargs)

# %% ../nbs/32_arng.ipynb 24
def try2arng(a) -> nparray:
    '''Attempts to convert `a` to a numpy array representing an arithmetic range starting at 0.
    
    Parameters
    ----------
    a : Any
        The input object to convert.
        
    Returns
    -------
    np.array
        An arithmetic range numpy array if possible, otherwise `a` itself.
    '''
    if isarng(a): return a
    try:
        arange = np.arange(_safe_len(a))
        return arange
    except: return a

# %% ../nbs/32_arng.ipynb 26
def arr2arng(a: indexlike) -> nparray:
    '''Converts an input array-like to a numpy array representing an arithmetic range starting at 0.
    
    Parameters
    ----------
    a : indexlike
        The array-like object to convert.
        
    Returns
    -------
    np.array
        A numpy array representing an arithmetic range.
    '''
    if isarng(a): return a
    arange = np.arange(len(asnp(a)))
    return arange

# %% ../nbs/32_arng.ipynb 28
def df2arng(df: dataframe) -> nparray:
    '''Converts a DataFrame to a numpy array representing an arithmetic range based on its index.
    
    Parameters
    ----------
    df : dataframe
        The DataFrame to convert.
        
    Returns
    -------
    np.array
        A numpy array representing an arithmetic range based on the DataFrame's index.
    '''
    if isarng(df): return df
    return arr2arng(_safe_idx(df))

# %% ../nbs/32_arng.ipynb 30
def asarng(a, *args: P.args, **kwargs: P.kwargs) -> nparray:
    '''Ensures `a` is represented as a numpy array of an arithmetic range, applying conversions as needed.
    
    Parameters
    ----------
    a : Any
        The input object to convert.
    *args : P.args
        Additional positional arguments for conversion functions.
    **kwargs : P.kwargs
        Additional keyword arguments for conversion functions.
        
    Returns
    -------
    np.array
        A numpy array representing an arithmetic range, or `a` if conversions are unsuccessful.
    '''    
    if isarng(a): return a
    return arngchain(a, (
            df2arng, 
            arr2arng,
            try2arng,            
        ), *args, **kwargs
    )

# %% ../nbs/32_arng.ipynb 32
@wraps(asarng, assigned=(__DOC__, __NAME__, __MODULE__, __ANNOTATIONS__))
def arange(a, *args: P.args, **kwargs: P.kwargs) -> nparray: 
    '''An alias for `asarng`, ensuring `a` is represented as or converted to a numpy array of an arithmetic range.'''
    return asarng(a, *args, **kwargs)



__version__ = "0.0.32"
# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00__init__.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/00__init__.ipynb 3
import itertools
from importlib import import_module
from types import ModuleType

def loadmod(mod: ModuleType):
    if hasattr(mod, '__all__'): return mod.__all__
    if hasattr(mod, '__module__'): return loadmod(import_module(mod.__module__))
    raise AttributeError(f"module {mod} has no attribute '__all__'")

# %% ../nbs/00__init__.ipynb 4
from asto import (
    cons, atyp, cast, nums, 
    time, sets, path,
    slcs, arrs, arng, 
    bend, tens, chns,
)

from .cons import *
from .atyp import *
from .cast import *
from .nums import *
from .time import *
from .sets import *
from .path import *

from .slcs import *
from .arrs import *
from .arng import *

from .bend import *
from .tens import *
from .chns import *

# %% ../nbs/00__init__.ipynb 5
__all__ = sorted(set(itertools.chain(*[
    loadmod(mod) for mod in (
        cons, atyp, cast, nums, 
        time, sets, path,
        slcs, arrs, arng, 
        bend, tens, chns,
    )
])))
